@@ -1,7 +1,7 @@
-/* $Id: osdep-hpux.c 2647 2011-12-09 16:37:29Z nicm $ */
+/* $OpenBSD$ */
 
 /*
- * Copyright (c) 2009 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2009 Nicholas Marriott <nicholas.marriott@gmail.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -18,18 +18,16 @@
 
 #include <sys/types.h>
 
-#include <event.h>
-
 #include "tmux.h"
 
 char *
-osdep_get_name(unused int fd, unused char *tty)
+osdep_get_name(__unused int fd, __unused char *tty)
 {
 	return (NULL);
 }
 
 char *
-osdep_get_cwd(pid_t pid)
+osdep_get_cwd(__unused int fd)
 {
 	return (NULL);
 }
